program FilterTest(PgmFile, FsmFile, KeywordFile, output);

{  Executable filename:  FILTER.EXE                              }
{                                                                }
{  Programmer:  Br. David Carlson                                }
{                                                                }
{  Date:  June 23, 1992                                          }
{                                                                }
{  Revised:  March 30, 1993                                      }
{                                                                }
{       This program is a test of a simple FSM-based tokenizer.  }
{  It reads the FILTER.PGM text file and breaks it into numbers, }
{  words, and the special items ; and :=.  See the accompanying  }
{  drawing of the finite state machine that specifies these      }
{  tokens, which are imagined as coming from a program written   }
{  in a simple Pascal-like language.  This program simply echo   }
{  prints each token, 1 per line, with the description of each.  }
{  Words are described as keywords or identifiers, with the list }
{  of possible keywords read in from the text file FILTER.KEY.   }
{  The program reads the text file FILTER.FSM for its data on    }
{  the finite state machine.                                     }


const
   MaxChars = 12;   { max # of chars in a word }
   MaxKeywords = 50;
   MaxStates = 10;
   MaxCategories = 15;

type
   ShortString = string[MaxChars];
   TokenType = record
                  item, category: ShortString
               end;
   TokenArray = array[1..MaxStates] of ShortString;
   CategoriesArray = array[1..MaxCategories] of ShortString;
   KeywordArray = array[1..MaxKeywords] of ShortString;
   TableType = array[1..MaxStates, 1..MaxCategories] of integer;

var
   token: TokenType;
   PgmFile, FsmFile, KeywordFile: text;
   SavedChar: char;   { holds a char to be saved for the next token }
   PushedBack: boolean;   { true will indicate that a char has been saved }
   NumKeywords, NumStates, NumCategories: integer;
   table: TableType;
   TokenAccepted: TokenArray;
   categories: CategoriesArray;
   keywords: KeywordArray;


procedure initialize(var SavedChar: char; var PushedBack: boolean);

{ Given:   nothing.                                               }
{ Task:    To initialize the global values used by PushBack.      }
{ Return:  SavedChar and PushedBack.                              }

   begin
      SavedChar := ' ';
      PushedBack := false
   end;


function IsDigit(ch: char): boolean;

{ Given:  A character ch.                                     }
{ Task:   Determine if this character represents a digit.     }
{ Return: True or false to indicate if ch is a digit.         }

   begin
      IsDigit := ((ch >= '0') and (ch <= '9'))
   end;


function IsAlpha(ch: char): boolean;

{ Given:  A character ch.                                     }
{ Task:   Determine if this character is a letter of alphabet.}
{ Return: True or false to indicate if ch is a letter.        }

   begin
      IsAlpha := (((ch >= 'a') and (ch <= 'z'))
                  or
                  ((ch >= 'A') and (ch <= 'Z')))
   end;


procedure GetChar(var PgmFile: text; var ch: char;
                  var status: ShortString);

{ Given:  the text input file Pgmfile and global values SavedChar    }
{         and PushedBack.                                            }
{ Task:   Find the next char to be processed.                        }
{ Return: That char as ch and status to indicate status of result.   }

   var
      classify: boolean;

   begin
      classify := false;
      if PushedBack then   { use char that was saved }
         begin
            PushedBack := false;
            ch := SavedChar;
            classify := true
         end
      else if eof(PgmFile) then
         status := 'eof'
      else if eoln(PgmFile) then
         begin
            status := 'whitespace';   { category includes ' ' and eoln }
            readln(PgmFile)   { skip over the eoln mark }
         end
      else
         begin
            {$I-}
            read(PgmFile, ch);
            {$I+}
            if IOResult <> 0 then
               begin
                  writeln('ERROR: unable to read a character from FILTER.PGM');
                  halt(1)
               end
            else
               classify := true
         end;

      if classify then
         begin
            if IsDigit(ch) then
               status := 'digit'
            else if IsAlpha(ch) then
               status := 'alpha'
            else if ch = ':' then
               status := 'colon'
            else if ch = '=' then
               status := 'equals'
            else if ch = ';' then
               status := 'semicolon'
            else if ch = ' ' then
               status := 'whitespace'
            else
               status := 'other'
         end
   end;


procedure PushBack(ch: char);

{ Given:  Char ch.                                           }
{ Task:   Save this char for future processing.              }
{ Return: In the GLOBAL values SavedChar and PushedBack the  }
{         char being saved and true to indicate that a char  }
{         has been saved.                                    }

   begin
      SavedChar := ch;
      PushedBack := true
   end;


procedure print(var token: TokenType);

{ Given:  The token to be printed.                            }
{ Task:   Print the token and its category.                   }
{ Return: nothing.                                            }

   begin
      with token do
         writeln(item:15, category:15)
   end;


procedure GetToken(var PgmFile: text; var token: TokenType);

{ Given:  Input text file PgmFile, which may be partially read.   }
{         Also uses the global (for speed) variable TokenAccepted.}
{ Task:   Read the next token from the file.                      }
{ Return: This token.                                             }

   var
      ch: char;
      status: ShortString;
      state, OldState, count: integer;


   function NextState(state: integer; var status: ShortString): integer;

   {  Given:   The current state and the status of the current input.      }
   {           Also uses the global variables (left as globals for speed): }
   {           NumStates, NumCategories, table, and categories.            }
   { Task:     Use the table to look up the next state.                    }
   { Return:   This next state.                                            }

      var
         column: integer;
         found: boolean;

      begin
         if (state < 1) or (state > NumStates) then
            NextState := -1   { error }
         else
            begin
               column := 1;
               found := false;
               while (column <= NumCategories) and (not found) do
                  if status = categories[column] then
                     found := true
                  else
                     column := column + 1;

               if found then
                  NextState := table[state, column]
               else
                  NextState := -1   { error }
            end
      end;


   procedure KeywordCheck(var token: TokenType);

   { Given:  A token, the array of keywords, and the number of keywords  }
   {         stored in this array.  All items except token are passed as }
   {         globals for speed.                                          }
   { Task:   If token is in the word category, adjust it to indicate     }
   {         whether it is an identifier or keyword.                     }
   { Return: Updated token with adjusted category field.                 }

      var
         k: integer;
         found: boolean;

      begin
         with token do
            begin
	       k := 1;
               found := false;
               while (k <= NumKeywords) and (not found) do
	          if keywords[k] = item then
	             found := true
	          else
   	             k := k + 1;
	       if found then
	          category := 'keyword'
	       else
	          category := 'identifier'
	    end
      end;


   begin   { procedure GetToken }
      count := 0;   { counts the number of chars read }
      state := 1;
      with token do
         begin
            item := '';   { initialize to empty string }
            category := '';
            repeat
               OldState := state;
               GetChar(PgmFile, ch, status);
               if (status <> 'whitespace') and (status <> 'eof') then
                  begin
                     count := count + 1;
                     if count <= MaxChars then   { discard chars beyond max }
                        begin
                           item[count] := ch;
                           item[0] := chr(count)   { record string length }
                        end
                  end;
               state := NextState(state, status)
            until (state = 0) or (state = -1);

            if state = 0 then   { a token has been successfully found }
               begin
                  if (status <> 'whitespace') and (status <> 'eof') then
                     begin   { we recorded 1 char too many in this token }
                        item[0] := chr(count - 1);
                        PushBack(ch)
                     end;
                  category := TokenAccepted[oldstate];
                  if category = 'word' then
                     KeywordCheck(token)
               end
            else   { state = -1 }
               category := 'error'
         end
   end;


procedure ReadFsmFile(var FsmFile: text; var NumStates, NumCategories:
   integer; var table: TableType; var TokenAccepted: TokenArray;
   var categories: CategoriesArray);

{ Given:   The text file FsmFile containing the appropriate data.    }
{ Task:    Read this data so as to determine the value returned:     }
{ Return:  The number of states, the number of categories, the       }
{          transition table, and the array of tokens accepted.       }

   var
      row, column, junk: integer;

   begin
      assign(FsmFile, 'FILTER.FSM');
      {$I-}
      reset(FsmFile);
      {$I+}
      if IOResult <> 0 then
         begin
            writeln('Cannot open FILTER.FSM file');
            halt(2)
         end;

      readln(FsmFile, NumStates);
      if (NumStates < 1) or (NumStates > MaxStates) then
         begin
            writeln('FILTER.FSM has invalid number of states: ', NumStates);
            halt(3)
         end;
      for row := 1 to NumStates do
         readln(FsmFile, TokenAccepted[row]);

      readln(FsmFile, NumCategories);
      if (NumCategories < 1) or (NumCategories > MaxCategories) then
         begin
            writeln('FILTER.FSM has invalid number of categories: ',
               NumCategories);
            halt(4)
         end;
      for column := 1 to NumCategories do
         readln(FsmFile, categories[column]);

      readln(FsmFile, junk);   { skip the comment line }
      for row := 1 to NumStates do
         for column := 1 to NumCategories do
            read(FsmFile, table[row, column])
   end;


procedure GetKeywords(var KeywordFile: text; var keywords: KeywordArray;
		      var NumKeywords: integer);

{ Given:  The text file called KeywordFile containing the keywords. }
{ Task:   Read the keywords into the array keywords.                }
{ Return: The keywords array and the count of the number of words.  }

   var
      k: integer;

   begin
      assign(KeywordFile, 'FILTER.KEY');
      {$I-}
      reset(KeywordFile);
      {$I+}
      if IOResult <> 0 then
         begin
            writeln('Cannot open FILTER.KEY');
	    halt(2)
         end;

      readln(KeywordFile, NumKeywords);
      if (NumKeywords < 0) or (NumKeywords > MaxKeywords) then
         begin
            writeln('Invalid number for the number of keywords');
            halt(3)
         end;

      for k := 1 to NumKeywords do
         readln(KeywordFile, keywords[k])
   end;


begin
   initialize(SavedChar, PushedBack);
   ReadFsmFile(FsmFile, NumStates, NumCategories,
      table, TokenAccepted, categories);
   GetKeywords(KeywordFile, keywords, NumKeywords);

   assign(PgmFile, 'FILTER.PGM');
   {$I-}
   reset(PgmFile);
   {$I+}
   if IOResult <> 0 then
      begin
         writeln('Cannot open FILTER.PGM file');
         halt(5)
      end;

   while not eof(PgmFile) do
      begin
         GetToken(PgmFile, token);
         print(token)
      end;

   { Let's try to get a token when at EOF: }
   GetToken(PgmFile, token);
   print(token);

   writeln('End of FilterTest program')
end.