program ParserTest(PgmFile, KeywordFile, FsmFile, TableFile, output);

{  Executable filename:  PARSER.EXE                             }
{                                                               }
{  Programmer:  Br. David Carlson                               }
{                                                               }
{  Date:  July 25, 1992                                         }
{                                                               }
{  Revised:  March 30, 1993                                     }
{                                                               }
{  Description:                                                 }
{       This program parses the program in the text file        }
{  PARSER.PGM, using as keywords the list in the text file      }
{  PARSER.KEY, according to the grammar rule information in     }
{  the text file PARSER.TBL.  The program prints on the screen  }
{  the parse tree and a success message if the program can be   }
{  parsed.  If not, the program prints a failure message along  }
{  with a partial parse tree.  The parse trees are printed in   }
{  indented form.                                               }
{                                                               }
{  Files used:                                                  }
{       PgmFile          PARSER.PGM     program to be parsed    }
{       KeywordFile      PARSER.KEY     list of keywords        }
{       FsmFile          PARSER.FSM     FSM tokenizer info      }
{       TableFile        PARSER.TBL     table of grammar info   }

const
   MaxChars = 12;
   MaxKeywords = 50;
   MaxStates = 10;
   MaxCategories = 15;
   MaxLookAheads = 16;
   MaxNonTerminals = 10;
   MaxNumRules = 16;
   MaxRuleLength = 8;
   tab = '   ';

type
   ShortString = string[MaxChars];
   TokenType = record
		  item, category: ShortString
	       end;
   TokenArray = array[1..MaxStates] of ShortString;
   CategoriesArray = array[1..MaxCategories] of ShortString;
   RuleBodyArray = array[1..MaxRuleLength] of TokenType;
   RuleType = record
                 length: integer;
                 body: RuleBodyArray
              end;
   KeywordArray = array[1..MaxKeywords] of ShortString;
   FsmTableType = array[1..MaxStates, 1..MaxCategories] of integer;
   LookAheadsType = array[1..MaxLookAheads] of TokenType;
   NonTerminalsType = array[1..MaxNonTerminals] of ShortString;
   RuleTableType = array[1..MaxNonTerminals, 1..MaxLookAheads] of integer;
   RuleArrayType = array[1..MaxNumRules] of RuleType;
   StackInfoType = record
                      token: TokenType;
                      NumTabs: integer
                   end;
   StackItemPtr = ^StackItemType;
   StackItemType = record
                      info: StackInfoType;
                      next: StackItemPtr
                   end;
   StackType = StackItemPtr;

var
   LookAheadSymbol: TokenType;
   keywords: KeywordArray;
   LookAheads: LookAheadsType;
   NonTerminals: NonTerminalsType;
   TokenAccepted: TokenArray;
   categories: CategoriesArray;
   RuleTable: RuleTableType;
   FsmTable: FsmTableType;
   rules: RuleArrayType;
   stack: StackType;
   StackValue, NewStackValue: StackInfoType;
   k, RuleNumber, NumKeywords, NumLookAheads, NumStates, NumCategories,
   NumNonTerminals, NumRules: integer;
   PgmFile, KeywordFile, FsmFile, TableFile: text;
   SavedChar: char;
   PushedBack: boolean;


procedure StackInitialize(var stack: StackType; value: ShortString);

{ Given:   value to place on stack.                              }
{ Task:    place this value on an otherwise empty stack.         }
{ Return:  this stack                                            }

   begin
      new(stack);
      with stack^ do
         begin
            info.token.item := value;
            info.token.category := 'nonterminal';
            info.NumTabs := 0;
            next := nil
         end
   end;


function empty(var stack: StackType): boolean;

   begin
      empty := (stack = nil)
   end;


procedure push(var stack: StackType; var value: StackInfoType);

{ Given:   current stack and value to push.                      }
{ Task:    push this value onto the stack.                       }
{ Return:  the updated stack.                                    }

   var
      p: StackItemPtr;

   begin
      new(p);
      with p^ do
         begin
            info := value;
            next := stack
         end;
      stack := p
   end;


procedure pop(var stack: StackType; var value: StackInfoType);

{ Given:   Current stack.                                        }
{ Task:    Pop 1 value from the stack.                           }
{ Return:  This value and the updated stack.                     }

   var
      p: StackItemPtr;

   begin
      if empty(stack) then
         begin
            writeln('Attempt to pop from an empty stack');
            halt(1)
         end
      else
         begin
            value := stack^.info;
            p := stack;
            stack := stack^.next;
            dispose(p)
         end
   end;


procedure SaveInitialize(var SavedChar: char; var PushedBack: boolean);

{ Given:   nothing.                                               }
{ Task:    To initialize the global values used by PushBack.      }
{ Return:  SavedChar and PushedBack.                              }

   begin
      SavedChar := ' ';
      PushedBack := false
   end;


procedure print(var value: StackInfoType);

{ Given:  The stack value to be printed.                      }
{ Task:   Print the value with indicated number of tabs.      }
{ Return: nothing.                                            }

   var
      k: integer;

   begin
      with value do
         begin
            for k := 1 to NumTabs do
               write(tab);
            write(token.item)
         end
   end;


procedure GetToken(var InFile: text; var token: TokenType);

{ Given:  Input text file InFile, which may be partially read.      }
{         Also uses the global (for speed) variable TokenAccepted.  }
{ Task:   Read the next token from the file.  Note that this        }
{         procedure distinguishes between keywords and identifiers. }
{ Return: This token.                                               }

   var
      ch: char;
      status: ShortString;
      state, OldState, count: integer;


   procedure GetChar(var PgmFile: text; var ch: char;
      var status: ShortString);

   { Given:  the text input file Pgmfile and global values SavedChar    }
   {         and PushedBack.                                            }
   { Task:   Find the next char to be processed.                        }
   { Return: That char as ch and status to indicate status of result.   }

      var
         classify: boolean;


      function IsDigit(ch: char): boolean;

      { Given:  A character ch.                                     }
      { Task:   Determine if this character represents a digit.     }
      { Return: True or false to indicate if ch is a digit.         }

         begin
            IsDigit := ((ch >= '0') and (ch <= '9'))
         end;


      function IsAlpha(ch: char): boolean;

      { Given:  A character ch.                                     }
      { Task:   Determine if this character is a letter of alphabet.}
      { Return: True or false to indicate if ch is a letter.        }

         begin
            IsAlpha := (((ch >= 'a') and (ch <= 'z'))
                        or
                        ((ch >= 'A') and (ch <= 'Z')))
         end;


      begin   { procedure GetChar }
         classify := false;
         if PushedBack then   { use char that was saved }
            begin
               PushedBack := false;
               ch := SavedChar;
               classify := true
            end
         else if eof(PgmFile) then
            status := 'eof'
         else if eoln(PgmFile) then
            begin
               status := 'whitespace';   { category includes ' ' and eoln }
               readln(PgmFile)   { skip over the eoln mark }
            end
         else
            begin
               {$I-}
               read(PgmFile, ch);
               {$I+}
               if IOResult <> 0 then
                  begin
                     writeln('ERROR: unable to read a character from PARSER.PGM');
                     halt(2)
                  end
               else
                  classify := true
            end;

         if classify then
            begin
               if IsDigit(ch) then
                  status := 'digit'
               else if IsAlpha(ch) then
                  status := 'alpha'
               else if ch = ':' then
                  status := 'colon'
               else if ch = '=' then
                  status := 'equals'
               else if ch = ';' then
                  status := 'semicolon'
               else if ch = ' ' then
                  status := 'whitespace'
               else
                  status := 'other'
            end
      end;


   procedure PushBack(ch: char);

   { Given:  Char ch.                                           }
   { Task:   Save this char for future processing.              }
   { Return: In the GLOBAL values SavedChar and PushedBack the  }
   {         char being saved and true to indicate that a char  }
   {         has been saved.                                    }

      begin
         SavedChar := ch;
         PushedBack := true
      end;


   function NextState(state: integer; var status: ShortString): integer;

   {  Given:   The current state and the status of the current input.      }
   {           Also uses the global variables (left as globals for speed): }
   {           NumStates, NumCategories, FsmTable, and categories.         }
   { Task:     Use the FsmTable to look up the next state.                 }
   { Return:   This next state.                                            }

      var
         column: integer;
         found: boolean;

      begin
         if (state < 1) or (state > NumStates) then
            NextState := -1   { error }
         else
            begin
               column := 1;
               found := false;
               while (column <= NumCategories) and (not found) do
                  if status = categories[column] then
                     found := true
                  else
                     column := column + 1;

               if found then
                  NextState := FsmTable[state, column]
               else
                  NextState := -1   { error }
            end
      end;


   procedure KeywordCheck(var token: TokenType);

   { Given:  A token, the array of keywords, and the number of keywords  }
   {         stored in this array.  All items except token are passed as }
   {         globals for speed.                                          }
   { Task:   If token is in the word category, adjust it to indicate     }
   {         whether it is an identifier or keyword.                     }
   { Return: Updated token with adjusted category field.                 }

      var
         k: integer;
         found: boolean;

      begin
         with token do
            begin
	       k := 1;
               found := false;
               while (k <= NumKeywords) and (not found) do
	          if keywords[k] = item then
	             found := true
	          else
   	             k := k + 1;
	       if found then
	          category := 'keyword'
	       else
	          category := 'identifier'
	    end
      end;


   begin   { procedure GetToken }
      count := 0;   { counts the number of chars read }
      state := 1;
      with token do
         begin
            item := '';   { initialize to empty string }
            category := '';
            repeat
               OldState := state;
               GetChar(PgmFile, ch, status);
               if (status <> 'whitespace') and (status <> 'eof') then
                  begin
                     count := count + 1;
                     if count <= MaxChars then   { discard chars beyond max }
                        begin
                           item[count] := ch;
                           item[0] := chr(count)   { record string length }
                        end
                  end;
               state := NextState(state, status)
            until (state = 0) or (state = -1);

            if state = 0 then   { a token has been successfully found }
               begin
                  if (status <> 'whitespace') and (status <> 'eof') then
                     begin   { we recorded 1 char too many in this token }
                        item[0] := chr(count - 1);
                        PushBack(ch)
                     end;
                  category := TokenAccepted[oldstate];
                  if category = 'word' then
                     KeywordCheck(token)
               end
            else   { state = -1 }
               category := 'error'
         end
   end;


procedure GetKeywords(var KeywordFile: text; var keywords: KeywordArray;
		      var NumKewords: integer);

{ Given:  The text file called KeywordFile containing the keywords.   }
{ Task:   Read the keywords into the array called keywords.           }
{ Return: The keywords array and the count of the number of keywords. }

   var
      k: integer;

   begin
      assign(KeywordFile, 'PARSER.KEY');
      {$I-}
      reset(KeywordFile);
      {$I+}
      if IOResult <> 0 then
         begin
            writeln('Cannot open PARSER.KEY');
	    halt(3)
         end;

      readln(KeywordFile, NumKeywords);
      if (NumKeywords < 0) or (NumKeywords > MaxKeywords) then
         begin
            writeln('Invalid number given for number of keywords');
            halt(4)
         end;

      for k := 1 to NumKeywords do
         readln(KeywordFile, keywords[k])
   end;


procedure GetTableInfo(var TableFile: text;
   var NumLookAheads, NumNonTerminals, NumRules: integer;
   var LookAheads: LookAheadsType; var NonTerminals: NonTerminalsType;
   var rules: RuleArrayType; var RuleTable: RuleTableType);

{ Given:   text file TableFile with info in required format          }
{ Task:    read this info into the other 8 parameters                }
{ Return:  this info                                                 }

   var
      BodyPart, TempString: ShortString;
      row, column: integer;
      NextCh: char;
      RecordNextCh: boolean;


   procedure ReadString(var f: text; var s: ShortString; var NextCh: char;
      RecordNextCh: boolean);

   { Given:   Text file f to read from and NextChar contains next char   }
   {          to be processed.  RecordNextCh indicates if NextCh is to   }
   {          included in string s.                                      }
   { Assumes: That this line of f contains at least 1 char to read.      }
   { Task:    Read string from f up to EOLN or first blank.              }
   { Return:  This string and the next char read after the string.       }

      var
         count : integer;
         continue: boolean;

      begin
         if RecordNextCh then
            begin
               count := 1;
               s[count] := NextCh;
               continue := (NextCh <> ' ')
            end
         else
            begin
               count := 0;
               continue := true
            end;

         while (not eoln(f)) and continue do
            begin
               read(f, NextCh);
               if count > MaxChars then
                  continue := false
               else if NextCh = ' ' then
                  continue := false
               else
                  begin
                     count := count + 1;
                     s[count] := NextCh
                  end
            end;

         s[0] := chr(count)   { record the string's length }
      end;


   procedure SkipBlanks(var f: text; var NextCh: char);

   { Given:   Text file f containing at least 1 char on the current line  }
   {          and NextChar containing 1 char already read from this line. }
   { Task:    Read until EOLN or first nonblank character is read.        }
   { Return:  This character as NextCh (or blank if no nonblank found)    }

      begin
         while (not eoln(f)) and (NextCh = ' ') do
            read(f, NextCh)
      end;


   begin   { procedure GetTableInfo }
      assign(TableFile, 'PARSER.TBL');
      {$I-}
      reset(TableFile);
      {$I+}
      if IOResult <> 0 then
         begin
            writeln('Cannot open PARSER.TBL');
            halt(5)
         end;

      readln(TableFile, NumLookAheads);
      for row := 1 to NumLookAheads do
         with LookAheads[row] do
            begin
               ReadString(TableFile, item, NextCh, false);
               SkipBlanks(TableFile, NextCh);
               ReadString(TableFile, category, NextCh, true);
               readln(TableFile)
            end;

      readln(TableFile, NumNonTerminals);
      for row := 1 to NumNonTerminals do
         begin
            ReadString(TableFile, NonTerminals[row], NextCh, false);
            for column := 1 to NumLookAheads do
               read(TableFile, RuleTable[row, column]);
            readln(TableFile)
         end;

      readln(TableFile, NumRules);
      for row := 1 to NumRules do
         begin
            ReadString(TableFile, TempString, NextCh, false); { skip index }
            SkipBlanks(TableFile, NextCh);
            ReadString(TableFile, TempString, NextCh, true); { skip rule head }
            SkipBlanks(TableFile, NextCh);
            ReadString(TableFile, TempString, NextCh, true); { skip the -> }
            with rules[row] do
               begin
                  read(TableFile, length);
                  NextCh := ' ';
                  for column := 1 to length do
                     begin
                        SkipBlanks(TableFile, NextCh);
                        ReadString(TableFile, BodyPart, NextCh, true);
                        with body[column] do
                           begin
                              if BodyPart[1] = '<' then
                                 category := 'nonterminal'
                              else
                                 category := 'terminal';
                              item := BodyPart
                           end
                     end
               end;
            readln(TableFile)
         end
   end;


procedure FindRule(var LookAheadSymbol: TokenType;
   var LookAheads: LookAheadsType; NumLookAheads: integer;
   var StackNonTerminal: ShortString;
   var NonTerminals: NonTerminalsType; NumNonTerminals: integer;
   var RuleTable: RuleTableType; var RuleNumber: integer);

{ Given:   LookAhead symbol to be found in LookAheads array which has  }
{          NumLookAheads items, StackNonTerminal to be found in the    }
{          NonTerminals array which has NumNonTerminals items.         }
{ Task:    To find the rule number in RuleTable corresponding to the   }
{          locations found in the above 2 arrays.  If either lookup    }
{          fails, use 0 for the rule number.                           }
{ Return:  This RuleNumber.                                            }

   var
      row, column: integer;
      found: boolean;

   begin
      column := 1;
      found := false;

      while (column <= NumLookAheads) and (not found) do
         if LookAheads[column].category = LookAheadSymbol.category then
            begin
               if (LookAheads[column].category = 'special') or
                  (LookAheads[column].category = 'keyword') then
                  begin
                     if LookAheads[column].item = LookAheadSymbol.item then
                        found := true
                     else
                        column := column + 1
                  end
               else
                  found := true
            end
         else
            column := column + 1;

      if not found then
         begin
            RuleNumber := 0;
            exit
         end;

      row := 1;
      found := false;

      while (row <= NumNonTerminals) and (not found) do
         if NonTerminals[row] = StackNonTerminal then
            found := true
         else
            row := row + 1;

      if not found then
         begin
            RuleNumber := 0;
            exit
         end;

      RuleNumber := RuleTable[row, column]
   end;


procedure ParseFailure;

{ Given:   nothing.                                          }
{ Task:    Print failure message and halt the program.       }
{ Return:  nothing.                                          }

   begin
      writeln;
      writeln('Parsing has failed!');
      halt(6)
   end;


procedure ReadFsmFile(var FsmFile: text; var NumStates, NumCategories:
   integer; var FsmTable: FsmTableType; var TokenAccepted: TokenArray;
   var categories: CategoriesArray);

{ Given:   The text file FsmFile containing the appropriate data.    }
{ Task:    Read this data so as to determine the value returned:     }
{ Return:  The number of states, the number of categories, the FSM   }
{          transition table, and the array of tokens accepted.       }

   var
      row, column, junk: integer;

   begin
      assign(FsmFile, 'PARSER.FSM');
      {$I-}
      reset(FsmFile);
      {$I+}
      if IOResult <> 0 then
         begin
            writeln('Cannot open PARSER.FSM file');
            halt(7)
         end;

      readln(FsmFile, NumStates);
      if (NumStates < 1) or (NumStates > MaxStates) then
         begin
            writeln('PARSER.FSM has invalid number of states: ', NumStates);
            halt(8)
         end;
      for row := 1 to NumStates do
         readln(FsmFile, TokenAccepted[row]);

      readln(FsmFile, NumCategories);
      if (NumCategories < 1) or (NumCategories > MaxCategories) then
         begin
            writeln('PARSER.FSM has invalid number of categories: ',
               NumCategories);
            halt(9)
         end;
      for column := 1 to NumCategories do
         readln(FsmFile, categories[column]);

      readln(FsmFile, junk);   { skip the comment line }
      for row := 1 to NumStates do
         for column := 1 to NumCategories do
            read(FsmFile, FsmTable[row, column])
   end;


begin
   GetKeywords(KeywordFile, keywords, NumKeywords);
   SaveInitialize(SavedChar, PushedBack);
   ReadFsmFile(FsmFile, NumStates, NumCategories, FsmTable,
      TokenAccepted, categories);
   GetTableInfo(TableFile, NumLookAheads, NumNonTerminals, NumRules,
      LookAheads, NonTerminals, rules, RuleTable);
   StackInitialize(stack, '<program>');

   assign(PgmFile, 'PARSER.PGM');
   {$I-}
   reset(PgmFile);
   {$I+}
   if IOResult <> 0 then
      begin
         writeln('Cannot open PARSER.PGM');
         halt(10)
      end;

   GetToken(PgmFile, LookAheadSymbol);

   while (not empty(stack)) do
      begin
         pop(stack, StackValue);
         print(StackValue);
         if StackValue.token.category = 'terminal' then
            begin
               write(tab);
               if LookAheadSymbol.category = 'EndMarker' then  { can't match }
                  ParseFailure
               else if (LookAheadSymbol.category = 'identifier') or
                  (LookAheadSymbol.category = 'number') then
                  begin
                     write(LookAheadSymbol.item);
                     if LookAheadSymbol.category = StackValue.token.item then
                        GetToken(PgmFile, LookAheadSymbol)
                     else
                         ParseFailure
                  end
               else if LookAheadSymbol.item = StackValue.token.item then
                  GetToken(PgmFile, LookAheadSymbol)
               else
                  ParseFailure;
               writeln
            end
         else   { StackValue.token.category is 'nonterminal' }
            begin
               writeln;
               FindRule(LookAheadSymbol, LookAheads, NumLookAheads,
                        StackValue.token.item, NonTerminals, NumNonTerminals,
                        RuleTable, RuleNumber);
               if RuleNumber = 0 then
                  ParseFailure
               else
                  with rules[RuleNumber] do
                     begin   { push parts of rule body in backwards order }
                        for k := length downto 1 do
                           begin
                              NewStackValue.NumTabs := StackValue.NumTabs + 1;
                              NewStackValue.token := body[k];
                              push(stack, NewStackValue)
                           end
                     end
            end
      end;

   if LookAheadSymbol.category = 'EndMarker' then
      writeln('Parsing has succeeded!')
   else
      ParseFailure
end.